# Generated by Django 3.0.3 on 2020-12-28 11:51

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('insuree', '0004_confirmationtype_education_profession_relation'),
        ('contract', '0004_auto_20201222_1424'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='contractdetails',
            name='policy_holder_insuree',
        ),
        migrations.RemoveField(
            model_name='historicalcontractdetails',
            name='policy_holder_insuree',
        ),
        migrations.AddField(
            model_name='contractdetails',
            name='insuree',
            field=models.ForeignKey(db_column='InsureeID', default=1, on_delete=django.db.models.deletion.DO_NOTHING, to='insuree.Insuree'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='historicalcontractdetails',
            name='insuree',
            field=models.ForeignKey(blank=True, db_column='InsureeID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='insuree.Insuree'),
        ),
        migrations.AlterField(
            model_name='contract',
            name='amendment',
            field=models.IntegerField(db_column='Amendment', default=0),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='amendment',
            field=models.IntegerField(db_column='Amendment', default=0),
        ),
    ]



import graphene
import graphene_django_optimizer as gql_optimizer

from django.db.models import Q
from core.schema import signal_mutation_module_before_mutating, OrderedDjangoFilterConnectionField
from core.utils import append_validity_filter
from contract.models import Contract, ContractDetails, \
    ContractContributionPlanDetails, ContractMutation
from contract.gql.gql_types import ContractGQLType, ContractDetailsGQLType, \
    ContractContributionPlanDetailsGQLType

from contract.gql.gql_mutations.contract_mutations import CreateContractMutation, \
    UpdateContractMutation, DeleteContractMutation, SubmitContractMutation, ApproveContractMutation, \
    ApproveContractBulkMutation, CounterContractMutation, \
    AmendContractMutation, RenewContractMutation, CounterContractBulkMutation, ContractCreateInvoiceBulkMutation
from contract.gql.gql_mutations.contract_details_mutations import CreateContractDetailsMutation, \
    UpdateContractDetailsMutation, DeleteContractDetailsMutation, \
    CreateContractDetailByPolicyHolderInsureeMutation
from contract.apps import ContractConfig
from contract.utils import filter_amount_contract


class Query(graphene.ObjectType):

    contract = OrderedDjangoFilterConnectionField(
        ContractGQLType,
        client_mutation_id=graphene.String(),
        insuree=graphene.UUID(),
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        amount_from=graphene.Decimal(),
        amount_to=graphene.Decimal(),
        applyDefaultValidityFilter=graphene.Boolean()
    )

    contract_details = OrderedDjangoFilterConnectionField(
        ContractDetailsGQLType,
        client_mutation_id=graphene.String(),
        orderBy=graphene.List(of_type=graphene.String),
    )

    contract_contribution_plan_details = OrderedDjangoFilterConnectionField(
        ContractContributionPlanDetailsGQLType,
        insuree=graphene.UUID(),
        contributionPlanBundle=graphene.UUID(),
        orderBy=graphene.List(of_type=graphene.String),
    )

    def resolve_contract(self, info, **kwargs):
        if not info.context.user.has_perms(ContractConfig.gql_query_contract_perms):
            if not info.context.user.has_perms(ContractConfig.gql_query_contract_policyholder_portal_perms):
                raise PermissionError("Unauthorized")

        filters = append_validity_filter(**kwargs)
        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        insuree = kwargs.get('insuree', None)
        if insuree:
            filters.append(Q(contractdetails__insuree__uuid=insuree))

        # amount filters
        amount_from = kwargs.get('amount_from', None)
        amount_to = kwargs.get('amount_to', None)
        if amount_from or amount_to:
            filters.append(filter_amount_contract(**kwargs))
        return gql_optimizer.query(Contract.objects.filter(*filters).all(), info)

    def resolve_contract_details(self, info, **kwargs):
        if not info.context.user.has_perms(ContractConfig.gql_query_contract_perms):
            if not info.context.user.has_perms(ContractConfig.gql_query_contract_policyholder_portal_perms):
                raise PermissionError("Unauthorized")

        filters = []
        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        return gql_optimizer.query(ContractDetails.objects.filter(*filters).all(), info)

    def resolve_contract_contribution_plan_details(self, info, **kwargs):
        if not info.context.user.has_perms(ContractConfig.gql_query_contract_perms):
            if not info.context.user.has_perms(ContractConfig.gql_query_contract_policyholder_portal_perms):
                raise PermissionError("Unauthorized")

        query = ContractContributionPlanDetails.objects.all()

        insuree = kwargs.get('insuree', None)
        contribution_plan_bundle = kwargs.get('contributionPlanBundle', None)

        if insuree:
            query = query.filter(
                contract_details__insuree__uuid=insuree
            )

        if contribution_plan_bundle:
            query = query.filter(
                contract_details__contribution_plan_bundle__id=contribution_plan_bundle
            )

        return gql_optimizer.query(query.all(), info)


class Mutation(graphene.ObjectType):
    create_contract = CreateContractMutation.Field()
    update_contract = UpdateContractMutation.Field()
    delete_contract = DeleteContractMutation.Field()
    submit_contract = SubmitContractMutation.Field()
    approve_contract = ApproveContractMutation.Field()
    approve_bulk_contract = ApproveContractBulkMutation.Field()
    counter_contract = CounterContractMutation.Field()
    counter_bulk_contract = CounterContractBulkMutation.Field()
    amend_contract = AmendContractMutation.Field()
    renew_contract = RenewContractMutation.Field()
    create_contract_invoice_bulk = ContractCreateInvoiceBulkMutation.Field()

    create_contract_details = CreateContractDetailsMutation.Field()
    update_contract_details = UpdateContractDetailsMutation.Field()
    delete_contract_details = DeleteContractDetailsMutation.Field()
    create_contract_details_by_ph_insuree = CreateContractDetailByPolicyHolderInsureeMutation.Field()


def on_contract_mutation(sender, **kwargs):
    uuids = kwargs['data'].get('uuids', [])
    if not uuids:
        uuid = kwargs['data'].get('uuid', None)
        uuids = [uuid] if uuid else []
    if not uuids:
        return []
    impacted_contracts = Contract.objects.filter(id__in=uuids).all()
    for contract in impacted_contracts:
        ContractMutation.objects.update_or_create(contract=contract, mutation_id=kwargs['mutation_log_id'])
    return []


signal_mutation_module_before_mutating["contract"].connect(on_contract_mutation)



# Generated by Django 3.0.3 on 2021-01-28 12:48

import core.models
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0009_mutationlog_client_mutation_details'),
        ('contract', '0010_auto_20210119_1204'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContractMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('contract', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='contract.Contract')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='contracts', to='core.MutationLog')),
            ],
            options={
                'db_table': 'contract_contractMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
    ]



import logging

from django.db import migrations

logger = logging.getLogger(__name__)


MIGRATION_SQL = """
    /* Contract */
    DECLARE @SystemRole INT
    SELECT @SystemRole = role.RoleID from tblRole role where IsSystem=64;
    /* Contract search*/
    IF NOT EXISTS (SELECT * FROM [tblRoleRight] WHERE [RoleID] = @SystemRole AND [RightID] = 152101)
    BEGIN
        INSERT [dbo].[tblRoleRight] ([RoleID], [RightID], [ValidityFrom]) 
        VALUES (@SystemRole, 152101, CURRENT_TIMESTAMP)
    END 
    /* Contract create*/
    IF NOT EXISTS (SELECT * FROM [tblRoleRight] WHERE [RoleID] = @SystemRole AND [RightID] = 152102)
    BEGIN
        INSERT [dbo].[tblRoleRight] ([RoleID], [RightID], [ValidityFrom]) 
        VALUES (@SystemRole, 152102, CURRENT_TIMESTAMP)
    END 
    /* Contract update*/
    IF NOT EXISTS (SELECT * FROM [tblRoleRight] WHERE [RoleID] = @SystemRole AND [RightID] = 152103)
    BEGIN
        INSERT [dbo].[tblRoleRight] ([RoleID], [RightID], [ValidityFrom]) 
        VALUES (@SystemRole, 152103, CURRENT_TIMESTAMP)
    END 
    /* Contract delete*/
    IF NOT EXISTS (SELECT * FROM [tblRoleRight] WHERE [RoleID] = @SystemRole AND [RightID] = 152104)
    BEGIN
        INSERT [dbo].[tblRoleRight] ([RoleID], [RightID], [ValidityFrom]) 
        VALUES (@SystemRole, 152104, CURRENT_TIMESTAMP)
    END 
    /* Contract renew*/
    IF NOT EXISTS (SELECT * FROM [tblRoleRight] WHERE [RoleID] = @SystemRole AND [RightID] = 152106)
    BEGIN
        INSERT [dbo].[tblRoleRight] ([RoleID], [RightID], [ValidityFrom]) 
        VALUES (@SystemRole, 152106, CURRENT_TIMESTAMP)
    END 
    /* Contract submit */
    IF NOT EXISTS (SELECT * FROM [tblRoleRight] WHERE [RoleID] = @SystemRole AND [RightID] = 152107)
    BEGIN
        INSERT [dbo].[tblRoleRight] ([RoleID], [RightID], [ValidityFrom]) 
        VALUES (@SystemRole, 152107, CURRENT_TIMESTAMP)
    END 
    /* payment approve */
    IF NOT EXISTS (SELECT * FROM [tblRoleRight] WHERE [RoleID] = @SystemRole AND [RightID] = 101408)
    BEGIN
        INSERT [dbo].[tblRoleRight] ([RoleID], [RightID], [ValidityFrom]) 
        VALUES (@SystemRole, 101408, CURRENT_TIMESTAMP)
    END 
"""


class Migration(migrations.Migration):
    dependencies = [
        ('contract', '0016_auto_20210208_1508')
    ]

    operations = [
        migrations.RunSQL(sql=MIGRATION_SQL),
    ]



from core import TimeUtils
from core.schema import OpenIMISMutation
from core.gql.gql_mutations import ObjectNotExistException
from contract.services import Contract as ContractService, \
    ContractDetails as ContractDetailsService, ContractToInvoiceService
from contract.models import Contract, ContractMutation
from contract.apps import ContractConfig
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError


class ContractCreateMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(ContractConfig.gql_mutation_create_contract_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        output = cls.create_contract(user=user, contract=data)
        if output["success"]:
            contract = Contract.objects.get(id=output["data"]["id"])
            ContractMutation.object_mutated(user, client_mutation_id=client_mutation_id, contract=contract)
            return None
        else:
            return f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def create_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.create(contract=contract)
        return output_data


class ContractUpdateMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(ContractConfig.gql_mutation_update_contract_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        output = cls.update_contract(user=user, contract=data)
        return None if output["success"] else f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def update_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.update(contract=contract)
        return output_data


class ContractDeleteMutationMixin:
    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _object_not_exist_exception(cls, obj_uuid):
        raise ObjectNotExistException(cls._model, obj_uuid)

    @classmethod
    def _validate_mutation(cls, user, **data):
        cls._validate_user(user)

    @classmethod
    def _validate_user(cls, user):
        if type(user) is AnonymousUser or not user.id:
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, uuid):
        output = cls.delete_contract(user=user, contract={"id": uuid})
        return None if output["success"] else f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def delete_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.delete(contract=contract)
        return output_data


class ContractSubmitMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(ContractConfig.gql_mutation_submit_contract_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        output = cls.submit_contract(user=user, contract=data)
        return None if output["success"] else f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def submit_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.submit(contract=contract)
        return output_data


class ContractApproveMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(ContractConfig.gql_mutation_approve_ask_for_change_contract_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        output = cls.approve_contract(user=user, contract=data)
        return None if output["success"] else f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def approve_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.approve(contract=contract)
        return output_data


class ContractCounterMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(ContractConfig.gql_mutation_approve_ask_for_change_contract_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        output = cls.counter_contract(user=user, contract=data)
        return None if output["success"] else f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def counter_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.counter(contract=contract)
        return output_data


class ContractAmendMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(ContractConfig.gql_mutation_amend_contract_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        output = cls.amend_contract(user=user, contract=data)
        if output["success"]:
            contract = Contract.objects.get(id=output["data"]["id"])
            ContractMutation.object_mutated(user, client_mutation_id=client_mutation_id, contract=contract)
            return None
        else:
            return f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def amend_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.amend(contract=contract)
        return output_data


class ContractRenewMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(ContractConfig.gql_mutation_renew_contract_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        output = cls.renew_contract(user=user, contract=data)
        if output["success"]:
            contract = Contract.objects.get(id=output["data"]["id"])
            ContractMutation.object_mutated(user, client_mutation_id=client_mutation_id, contract=contract)
            return None
        else:
            return f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def renew_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.renew(contract=contract)
        return output_data


class ContractDetailsFromPHInsureeMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(ContractConfig.gql_mutation_update_contract_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        output = cls.create_cd_from_ph_insuree(user=user, data=data)
        return None if output["success"] else f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def create_cd_from_ph_insuree(cls, user, data):
        contract_details_service = ContractDetailsService(user=user)
        data_contract = {
            "id": data["contract_id"]
        }
        data_insuree = {
            "id": data["policy_holder_insuree_id"]
        }
        output_data = contract_details_service.ph_insuree_to_contract_details(
            contract=data_contract,
            ph_insuree=data_insuree
        )
        return output_data


class ContractCreateInvoiceMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                ContractConfig.gql_invoice_create_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        output = cls.create_contract_invoice(user=user, data=data)
        if output["success"]:
            return None
        else:
            return f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def create_contract_invoice(cls, user, data):
        queryset = Contract.objects.filter(id=data['id'])
        if queryset.count() == 1:
            contract = queryset.first()
            contract_to_invoice_service = ContractToInvoiceService(user=user)
            output_data = contract_to_invoice_service.create_invoice(instance=contract, convert_to='InvoiceLine', user=user)
            return output_data



import datetime
import numbers
import base64
from unittest import mock
from django.test import TestCase
from uuid import UUID

import graphene
from contract.tests.helpers import *
from contract.models import Contract, ContractDetails
from core.test_helpers import create_test_technical_user
from policyholder.tests.helpers import *
from contribution_plan.tests.helpers import create_test_contribution_plan, \
    create_test_contribution_plan_bundle, create_test_contribution_plan_bundle_details
from contribution_plan.models import ContributionPlan, ContributionPlanBundle, ContributionPlanBundleDetails
from policyholder.models import PolicyHolder, PolicyHolderInsuree
from contract import schema as contract_schema
from graphene import Schema
from graphene.test import Client


class MutationTestContract(TestCase):

    class BaseTestContext:
        def __init__(self, user):
            self.user = user

    class AnonymousUserContext:
        user = mock.Mock(is_anonymous=True)

    @classmethod
    def setUpClass(cls):
        if not TechnicalUser.objects.filter(username='admin').exists():
            create_test_technical_user(username='admin', password='S\/pe®Pąßw0rd™', super_user=True)
        cls.user = User.objects.filter(username='admin').first()
        # some test data so as to created contract properly
        cls.income = 500
        cls.rate = 5
        cls.number_of_insuree = 2
        cls.policy_holder = create_test_policy_holder()
        cls.policy_holder2 = create_test_policy_holder()
        # create contribution plans etc
        cls.contribution_plan_bundle = create_test_contribution_plan_bundle()
        cls.contribution_plan = create_test_contribution_plan(
            custom_props={"json_ext": '{"calculation_rule":{"rate": "' + str(cls.rate) + '"}}'}
        )
        cls.contribution_plan_bundle_details = create_test_contribution_plan_bundle_details(
            contribution_plan=cls.contribution_plan,
            contribution_plan_bundle=cls.contribution_plan_bundle
        )
        from core import datetime
        # create policy holder insuree for that test policy holder
        for i in range(0, cls.number_of_insuree):
            ph_insuree = create_test_policy_holder_insuree(
                policy_holder=cls.policy_holder,
                contribution_plan_bundle=cls.contribution_plan_bundle,
                custom_props={
                    "last_policy": None,
                    "json_ext": '{"calculation_rule":{"income": "' + str(cls.income) + '"}}'
                }
            )
            create_test_policy(
                cls.contribution_plan.benefit_plan,
                ph_insuree.insuree,
                custom_props={
                    "start_date": datetime.datetime(2016, 3, 1),
                    "expiry_date": datetime.datetime(2021, 7, 1)
                }
            )

        cls.policy_holder_insuree = create_test_policy_holder_insuree(
            policy_holder=cls.policy_holder,
            contribution_plan_bundle=cls.contribution_plan_bundle
        )
        cls.policy_holder_insuree2 = create_test_policy_holder_insuree(
            policy_holder=cls.policy_holder,
            contribution_plan_bundle=cls.contribution_plan_bundle
        )

        cls.schema = Schema(
            query=contract_schema.Query,
            mutation=contract_schema.Mutation
        )
        cls.graph_client = Client(cls.schema)

    @classmethod
    def tearDownClass(cls):
        # tear down the test data created during set up
        PolicyHolderInsuree.objects.filter(policy_holder_id=str(cls.policy_holder.id)).delete()
        PolicyHolder.objects.filter(id=cls.policy_holder.id).delete()
        ContributionPlanBundleDetails.objects.filter(id=cls.contribution_plan_bundle_details.id).delete()
        ContributionPlanBundle.objects.filter(id=cls.contribution_plan_bundle.id).delete()
        ContributionPlan.objects.filter(id=cls.contribution_plan.id).delete()

    def test_mutation_contract_create_without_policy_holder(self):
        time_stamp = datetime.datetime.now()
        input_param = {
            "code": "XYZ:"+str(time_stamp),
        }
        self.add_mutation("createContract", input_param)
        result = self.find_by_exact_attributes_query(
            "contract",
            params=input_param,
        )["edges"]
        converted_id = base64.b64decode(result[0]['node']['id']).decode('utf-8').split(':')[1]
        # tear down the test data
        Contract.objects.filter(id=str(converted_id)).delete()
        self.assertEqual(
            (
                "XYZ:"+str(time_stamp),
            )
            ,
            (
                result[0]['node']['code'],
            )
        )

    def test_mutation_contract_create_with_policy_holder(self):
        time_stamp = datetime.datetime.now()
        input_param = {
            "code": "XYZ:"+str(time_stamp),
            "policyHolderId": str(self.policy_holder.id)
        }
        self.add_mutation("createContract", input_param)
        result = self.find_by_exact_attributes_query(
            "contract",
            params=input_param,
        )["edges"]
        converted_id = base64.b64decode(result[0]['node']['id']).decode('utf-8').split(':')[1]
        # tear down the test data
        ContractDetails.objects.filter(contract_id=str(converted_id)).delete()
        Contract.objects.filter(id=str(converted_id)).delete()

        self.assertEqual(
            (
                "XYZ:"+str(time_stamp),
            )
            ,
            (
                result[0]['node']['code'],
            )
        )

    def find_by_id_query(self, query_type, id, context=None):
        query = F'''
        {{
            {query_type}(id:"{id}") {{
                totalCount
                edges {{
                  node {{
                    id
                    version
                  }}
                  cursor
                }}
          }}
        }}
        '''

        query_result = self.execute_query(query, context=context)
        records = query_result[query_type]['edges']

        if len(records) > 1:
            raise ValueError(F"Ambiguous id {id} for query {query_type}")

        return records

    def find_by_exact_attributes_query(self, query_type, params, context=None):
        if "dateValidFrom" in params:
            params.pop('dateValidFrom')
        if "dateValidTo" in params:
            params.pop('dateValidTo')
        if "policyHolderId" in params:
            params.pop('policyHolderId')
        if "contributionPlanBundleId" in params:
            params.pop('contributionPlanBundleId')
        if "insureeId" in params:
            params.pop('insureeId')
        if "uuids" in params:
            params["id"] = params["uuids"][0]
            params.pop("uuids")
        node_content_str = "\n".join(params.keys()) if query_type == "contract" else ''
        query = F'''
        {{
            {query_type}({ 'contract_Id: "'+str(params["contractId"])+'", orderBy: ["-dateCreated"]'  if "contractId" in params else self.build_params(params)}) {{
                totalCount
                edges {{
                  node {{
                    id
                    {node_content_str}
                    version
                    {'amountDue' if query_type =='contract' else '' }
                    {'amountNotified' if query_type =='contract' else '' }
                    {'amountRectified' if query_type =='contract' else '' }
                    {'state' if query_type =='contract' else '' }
                    {'paymentReference' if query_type =='contract' else '' }
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query, context=context)
        records = query_result[query_type]
        return records

    def execute_query(self, query, context=None):
        if context is None:
            context = self.BaseTestContext(self.user)

        query_result = self.graph_client.execute(query, context=context)
        query_data = query_result['data']
        return query_data

    def add_mutation(self, mutation_type, input_params, context=None):
        mutation = f'''
        mutation 
        {{
            {mutation_type}(input: {{
               {self.build_params(input_params)}
            }})  

          {{
            internalId
            clientMutationId
          }}
        }}
        '''
        mutation_result = self.execute_mutation(mutation, context=context)
        return mutation_result

    def execute_mutation(self, mutation, context=None):
        if context is None:
            context = self.BaseTestContext(self.user)

        mutation_result = self.graph_client.execute(mutation, context=context)
        return mutation_result

    def build_params(self, params):
        def wrap_arg(v):
            if isinstance(v, str):
                return F'"{v}"'
            if isinstance(v, list):
                return json.dumps(v)
            if isinstance(v, bool):
                return str(v).lower()
            if isinstance(v, datetime.date):
                return graphene.DateTime.serialize(
                    datetime.datetime.fromordinal(v.toordinal()))
            return v

        params_as_args = [f'{k}:{wrap_arg(v)}' for k, v in params.items()]
        return ", ".join(params_as_args)



CONTRACT_STATE = {
    "ContractState": [
        {
            "value": "1",
            "label":
                {
                    "fr": "Demande d'information",
                    "en": "Request for information"
                }
        },
        {
            "value": "2",
            "label":
                {
                    "fr": "Brouillon",
                    "en": "Draft"
                }
        },
        {
            "value": "3",
            "label":
                {
                    "fr": "Offre",
                    "en": "offer"
                }
        },
        {
            "value": "4",
            "label":
                {
                    "fr": "En negociation",
                    "en": "Negotiable"

                }
        },
        {
            "value": "5",
            "label":
                {
                    "fr": "Apprové",
                    "en": "executable"

                }
        },
        {
            "value": "6",
            "label":
                {
                    "fr": "addendum",
                    "en": "addendum"
                }
        },
        {
            "value": "7",
            "label":
                {
                    "fr": "En cours",
                    "en": "effective"
                }
        },
        {
            "value": "8",
            "label":
                {
                    "fr": "Appliqué",
                    "en": "executed"

                }
        },
        {
            "value": "9",
            "label":
                {
                    "fr": "Suspendu",
                    "en": "Disputed"
                }
        },
        {
            "value": "10",
            "label":
                {
                    "fr": "Terminé",
                    "en": "terminated"
                }
        },
        {
            "value": "11",
            "label":
                {
                    "fr": "révision demandé",
                    "en": "counter"
                }
        }]
}


def get_message_approved_contract(code, name, contact_name, due_amount, payment_reference, language='en'):
    message_payment_notification = {
        "payment_notification":
            {"en":
                 F"""
                 Dear {contact_name} 
                 
                 The contract {code} - {name} was approved.
                 Please proceed to the payment of {due_amount} with the reference {payment_reference}. 
                 
                 Best regards, 
                 """
                ,
             "fr":
                 F"""
                 Monsieur, Madame {contact_name}
                 
                 le contract {code} - {name} à été apprové.
                 Veuillez faire un paiement de < CONTRACT - DUEAMOUNT > avec la référence {payment_reference}.
                 
                 Meilleurs Salutations 
                 """
             }
    }
    return message_payment_notification["payment_notification"][language]


def get_message_counter_contract(code, name, contact_name, language='en'):
    message_payment_notification = {
        "payment_notification":
            {"en":
                 F"""
                 Dear {contact_name} 

                 The contract {code} - {name} was counter.
                 Please proceed recheck the information and correct the issues, in case of questions please check contact us. 

                 Best regards, 
                 """
                ,
             "fr":
                 F"""
                 Monsieur, Madame {contact_name}

                 le contract {code} - {name} à été contré.
                 Veuillez verifier les information saisie, en cas de question veuillez nous contacter.

                 Meilleurs Salutations 
                 """
             }
    }
    return message_payment_notification["payment_notification"][language]



from core.gql.gql_mutations import DeleteInputType, mutation_on_uuids_from_filter_business_model
from core.gql.gql_mutations.base_mutation  import BaseMutation, BaseDeleteMutation
from .mutations import ContractCreateMutationMixin, ContractUpdateMutationMixin, \
    ContractDeleteMutationMixin, ContractSubmitMutationMixin, \
    ContractApproveMutationMixin, ContractCounterMutationMixin, \
    ContractAmendMutationMixin, ContractRenewMutationMixin, ContractCreateInvoiceMutationMixin
from contract.models import Contract
from contract.gql.gql_types import ContractGQLType
from contract.gql.gql_mutations.input_types import ContractCreateInputType, ContractUpdateInputType, \
    ContractSubmitInputType, ContractApproveInputType, ContractCounterInputType, \
    ContractApproveBulkInputType, ContractAmendInputType, ContractRenewInputType, \
    ContractCounterBulkInputType, ContractCreateInvoiceBulkInputType
from contract.tasks import approve_contracts, counter_contracts, create_invoice_from_contracts


class CreateContractMutation(ContractCreateMutationMixin, BaseMutation):
    _mutation_class = "CreateContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractCreateInputType):
        pass


class UpdateContractMutation(ContractUpdateMutationMixin, BaseMutation):
    _mutation_class = "UpdateContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractUpdateInputType):
        pass


class DeleteContractMutation(ContractDeleteMutationMixin, BaseDeleteMutation):
    _mutation_class = "DeleteContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(DeleteInputType):
        pass


class SubmitContractMutation(ContractSubmitMutationMixin, BaseMutation):
    _mutation_class = "SubmitContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractSubmitInputType):
        pass


class ApproveContractMutation(ContractApproveMutationMixin, BaseMutation):
    _mutation_class = "ApproveContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractApproveInputType):
        pass


class ApproveContractBulkMutation(ContractApproveMutationMixin, BaseMutation):
    _mutation_class = "ApproveContractBulkMutation"
    _mutation_module = "contract"
    _model = Contract

    @classmethod
    @mutation_on_uuids_from_filter_business_model(Contract, ContractGQLType, 'extended_filters', {})
    def async_mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        if "contract_uuids" in data or "uuids" in data:
            cls.approve_contracts(user=user, contracts=data)
        return None

    @classmethod
    def approve_contracts(cls, user, contracts):
        if "uuids" in contracts:
            contracts["uuids"] = list(contracts["uuids"].values_list("id", flat=True))
            approve_contracts.delay(user_id=f'{user.id}', contracts=contracts["uuids"])
        else:
            if "contract_uuids" in contracts:
                approve_contracts.delay(user_id=f'{user.id}', contracts=contracts["contract_uuids"])

    class Input(ContractApproveBulkInputType):
        pass


class CounterContractMutation(ContractCounterMutationMixin, BaseMutation):
    _mutation_class = "CounterContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractCounterInputType):
        pass


class ContractCreateInvoiceBulkMutation(ContractCreateInvoiceMutationMixin, BaseMutation):
    _mutation_class = "ContractCreateInvoiceBulkMutation"
    _mutation_module = "contract"
    _model = Contract

    @classmethod
    @mutation_on_uuids_from_filter_business_model(Contract, ContractGQLType, 'extended_filters', {})
    def async_mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        if "contract_uuids" in data or "uuids" in data:
            cls.create_contract_invoice(user=user, contracts=data)
        return None

    @classmethod
    def create_contract_invoice(cls, user, contracts):
        if "uuids" in contracts:
            contracts["uuids"] = list(contracts["uuids"].values_list("id", flat=True))
            create_invoice_from_contracts.delay(user_id=f'{user.id}', contracts=contracts["uuids"])
        else:
            if "contract_uuids" in contracts:
                create_invoice_from_contracts.delay(user_id=f'{user.id}', contracts=contracts["contract_uuids"])

    class Input(ContractCreateInvoiceBulkInputType):
        pass


class CounterContractBulkMutation(ContractCounterMutationMixin, BaseMutation):
    _mutation_class = "CounterContractBulkMutation"
    _mutation_module = "contract"
    _model = Contract

    @classmethod
    @mutation_on_uuids_from_filter_business_model(Contract, ContractGQLType, 'extended_filters', {})
    def async_mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        if "contract_uuids" in data or "uuids" in data:
            cls.counter_contracts(user=user, contracts=data)
        return None

    @classmethod
    def counter_contracts(cls, user, contracts):
        if "uuids" in contracts:
            contracts["uuids"] = list(contracts["uuids"].values_list("id", flat=True))
            counter_contracts.delay(user_id=f'{user.id}', contracts=contracts["uuids"])
        else:
            if "contract_uuids" in contracts:
                counter_contracts.delay(user_id=f'{user.id}', contracts=contracts["contract_uuids"])

    class Input(ContractCounterBulkInputType):
        pass


class AmendContractMutation(ContractAmendMutationMixin, BaseMutation):
    _mutation_class = "AmendContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractAmendInputType):
        pass


class RenewContractMutation(ContractRenewMutationMixin, BaseMutation):
    _mutation_class = "RenewContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractRenewInputType):
        pass


# Generated by Django 3.0.3 on 2020-12-21 13:12

import core.fields
import datetime
import dirtyfields.dirtyfields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import jsonfallback.fields
import simple_history.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('policyholder', '0004_auto_20201214_1302'),
        ('contribution_plan', '0004_auto_20201217_0946'),
        ('policy', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Contract',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', models.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', models.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', models.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', models.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('amount_notified', models.FloatField(db_column='AmountNotified')),
                ('amount_rectified', models.FloatField(db_column='AmountRectified')),
                ('amount_due', models.FloatField(db_column='AmountDue')),
                ('date_approved', core.fields.DateTimeField(db_column='DateApproved')),
                ('date_payment_due', core.fields.DateField(db_column='DatePaymentDue')),
                ('state', models.SmallIntegerField(db_column='State')),
                ('payment_reference', models.CharField(db_column='PaymentReference', max_length=255)),
                ('policy_holder', models.ForeignKey(db_column='PolicyHolderUUID', on_delete=django.db.models.deletion.DO_NOTHING, to='policyholder.PolicyHolder')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='contract_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='contract_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblContract',
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ContractDetails',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', models.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', models.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('json_param', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_param', null=True)),
                ('contract', models.ForeignKey(db_column='ContractUUID', on_delete=django.db.models.deletion.DO_NOTHING, to='contract.Contract')),
                ('contribution_plan_bundle', models.ForeignKey(db_column='ContributionPlanBundleUUID', on_delete=django.db.models.deletion.DO_NOTHING, to='contribution_plan.ContributionPlanBundle')),
                ('policy_holder_insuree', models.ForeignKey(db_column='PolicyHolderInsureeUUID', on_delete=django.db.models.deletion.DO_NOTHING, to='policyholder.PolicyHolderInsuree')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='contractdetails_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='contractdetails_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblContractDetails',
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalContractDetails',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', models.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', models.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('json_param', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_param', null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('contract', models.ForeignKey(blank=True, db_column='ContractUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contract.Contract')),
                ('contribution_plan_bundle', models.ForeignKey(blank=True, db_column='ContributionPlanBundleUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contribution_plan.ContributionPlanBundle')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('policy_holder_insuree', models.ForeignKey(blank=True, db_column='PolicyHolderInsureeUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='policyholder.PolicyHolderInsuree')),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical contract details',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalContractContributionPlanDetails',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', models.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', models.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', models.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', models.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('contract_details', models.ForeignKey(blank=True, db_column='ContractDetailsUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contract.ContractDetails')),
                ('contribution_plan', models.ForeignKey(blank=True, db_column='ContributionPlanUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contribution_plan.ContributionPlan')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('policy', models.ForeignKey(blank=True, db_column='PolicyUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='policy.Policy')),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical contract contribution plan details',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalContract',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', models.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', models.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', models.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', models.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('amount_notified', models.FloatField(db_column='AmountNotified')),
                ('amount_rectified', models.FloatField(db_column='AmountRectified')),
                ('amount_due', models.FloatField(db_column='AmountDue')),
                ('date_approved', core.fields.DateTimeField(db_column='DateApproved')),
                ('date_payment_due', core.fields.DateField(db_column='DatePaymentDue')),
                ('state', models.SmallIntegerField(db_column='State')),
                ('payment_reference', models.CharField(db_column='PaymentReference', max_length=255)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('policy_holder', models.ForeignKey(blank=True, db_column='PolicyHolderUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='policyholder.PolicyHolder')),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical contract',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='ContractContributionPlanDetails',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', models.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', models.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', models.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', models.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('contract_details', models.ForeignKey(db_column='ContractDetailsUUID', on_delete=django.db.models.deletion.DO_NOTHING, to='contract.ContractDetails')),
                ('contribution_plan', models.ForeignKey(db_column='ContributionPlanUUID', on_delete=django.db.models.deletion.DO_NOTHING, to='contribution_plan.ContributionPlan')),
                ('policy', models.ForeignKey(db_column='PolicyUUID', on_delete=django.db.models.deletion.DO_NOTHING, to='policy.Policy')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='contractcontributionplandetails_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='contractcontributionplandetails_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblContractContributionPlanDetails',
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
    ]



from functools import lru_cache
from django.test import TestCase

from .helpers import *
from ..models import ContractContributionPlanDetails


class HelpersTest(TestCase):
    """
    Class to check whether the helper methods responsible for creating test data work correctly.
    """

    def test_create_test_contract(self):
        contract = self.__create_test_contract()
        db_contract = Contract.objects.filter(id=contract.uuid).first()
        self.assertEqual(db_contract, contract, "Failed to create contract in helper")

    def test_create_test_contract_custom(self):
        contract = self.__create_test_contract(custom=True)
        db_contract = Contract.objects.filter(id=contract.uuid).first()
        params = self.__custom_contract_params
        self.assertEqual(db_contract.version, params['version'])
        self.assertEqual(db_contract.policy_holder, params['policy_holder'])
        self.assertEqual(db_contract.state, params['state'])

    def test_create_test_contract_details(self):
        contract_details = self.__create_test_contract_details()
        db_contract_details = ContractDetails.objects.filter(id=contract_details.uuid).first()
        self.assertEqual(db_contract_details, contract_details, "Failed to create contract details in helper")

    def test_create_test_contract_details_custom(self):
        contract_details = self.__create_test_contract_details(custom=True)
        db_contract_details = ContractDetails.objects.filter(id=contract_details.uuid).first()
        params = self.__custom_contract_details_params

        self.assertEqual(db_contract_details.contract, params['contract'])
        self.assertEqual(db_contract_details.insuree, params['insuree'])
        self.assertEqual(db_contract_details.contribution_plan_bundle, params['contribution_plan_bundle'])

    def test_create_test_contract_contribution_plan_details(self):
        contract_contribution_plan_details = self.__create_test_contract_contribution_plan_details()
        db_contract_contribution_plan_details = ContractContributionPlanDetails.objects\
            .filter(id=contract_contribution_plan_details.uuid).first()
        self.assertEqual(db_contract_contribution_plan_details, 
                         contract_contribution_plan_details,
                         "Failed to create contract contribution plan details in helper")

    def test_create_test_contract_contribution_plan_details_custom(self):
        contract_contribution_plan_details = self.__create_test_contract_contribution_plan_details(custom=True)
        db_contract_contribution_plan_details = ContractContributionPlanDetails.objects\
            .filter(id=contract_contribution_plan_details.uuid).first()
        params = self.__custom_contract_contribution_plan_details_params
        self.assertEqual(db_contract_contribution_plan_details.contribution_plan, params['contribution_plan'])
        self.assertEqual(db_contract_contribution_plan_details.contract_details, params['contract_details'])

    @property
    @lru_cache(maxsize=2)
    def __custom_contract_params(self):
        return {
            'version': 2,
            'policy_holder': create_test_policy_holder(custom_props={'version': 2}),
            'state': 1,
        }

    @property
    @lru_cache(maxsize=2)
    def __custom_contract_details_params(self):
        return {
            'contract': self.__create_test_contract(True),
            'contribution_plan_bundle': create_test_contribution_plan_bundle(custom_props={'version': 3}),
            'insuree': create_test_insuree()
        }

    @property
    @lru_cache(maxsize=2)
    def __custom_contract_contribution_plan_details_params(self):
        return {
            'contribution_plan': create_test_contribution_plan(custom_props={'version': 2}),
            'contract_details': self.__create_test_contract_details(True),
        }

    def __create_test_instance(self, function, **kwargs):
        if kwargs:
            return function(**kwargs)
        else:
            return function()

    def __create_test_contract(self, custom=False):
        custom_params = self.__custom_contract_params if custom else {}
        return self.__create_test_instance(create_test_contract, custom_props=custom_params)

    def __create_test_contract_details(self, custom=False):
        custom_params = self.__custom_contract_details_params if custom else {}
        return self.__create_test_instance(create_test_contract_details, custom_props=custom_params)

    def __create_test_contract_contribution_plan_details(self, custom=False):
        custom_params = self.__custom_contract_contribution_plan_details_params if custom else {}
        return self.__create_test_instance(create_test_contract_contribution_plan_details,
                                           custom_props=custom_params)



import graphene

from core.schema import OpenIMISMutation, TinyInt


class ContractCreateInputType(OpenIMISMutation.Input):
    code = graphene.String(required=True, max_length=32)
    policy_holder_id = graphene.UUID(required=False)

    amount_notified = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    amount_rectified = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    amount_due = graphene.Decimal(max_digits=18, decimal_places=2, required=False)

    date_approved = graphene.DateTime(required=False)
    date_payment_due = graphene.Date(required=False)

    payment_reference = graphene.String(required=False)
    amendment = graphene.Int(required=False)

    date_valid_from = graphene.Date(required=False)
    date_valid_to = graphene.Date(required=False)
    json_ext = graphene.types.json.JSONString(required=False)


class ContractUpdateInputType(OpenIMISMutation.Input):
    id = graphene.UUID(required=True)
    code = graphene.String(required=False, max_length=32)
    policy_holder_id = graphene.UUID(required=False)

    amount_notified = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    amount_rectified = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    amount_due = graphene.Decimal(max_digits=18, decimal_places=2, required=False)

    date_approved = graphene.DateTime(required=False)
    date_payment_due = graphene.Date(required=False)

    payment_reference = graphene.String(required=False)
    amendment = graphene.Int(required=False)

    date_valid_from = graphene.Date(required=False)
    date_valid_to = graphene.Date(required=False)
    json_ext = graphene.types.json.JSONString(required=False)


class ContractSubmitInputType(OpenIMISMutation.Input):
    id = graphene.UUID(required=True)


class ContractApproveInputType(OpenIMISMutation.Input):
    id = graphene.UUID(required=True)


class ContractApproveBulkInputType(OpenIMISMutation.Input):
    contract_uuids = graphene.List(graphene.UUID, required=True)
    extended_filters = graphene.String(required=False)


class ContractCounterBulkInputType(OpenIMISMutation.Input):
    contract_uuids = graphene.List(graphene.UUID, required=True)
    extended_filters = graphene.String(required=False)


class ContractCounterInputType(OpenIMISMutation.Input):
    id = graphene.UUID(required=True)


class ContractCreateInvoiceBulkInputType(OpenIMISMutation.Input):
    contract_uuids = graphene.List(graphene.UUID, required=True)
    extended_filters = graphene.String(required=False)


class ContractAmendInputType(OpenIMISMutation.Input):
    id = graphene.UUID(required=True)
    amount_notified = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    amount_rectified = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    amount_due = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    date_approved = graphene.DateTime(required=False)
    date_payment_due = graphene.Date(required=False)
    payment_reference = graphene.String(required=False)
    date_valid_to = graphene.Date(required=False)
    json_ext = graphene.types.json.JSONString(required=False)


class ContractRenewInputType(OpenIMISMutation.Input):
    id = graphene.UUID(required=True)


class ContractDetailsCreateInputType(OpenIMISMutation.Input):
    id = graphene.UUID(required=False)
    contract_id = graphene.UUID(required=True)
    insuree_id = graphene.Int(required=True)
    contribution_plan_bundle_id = graphene.UUID(required=True)
    json_ext = graphene.types.json.JSONString(required=False)
    json_param = graphene.types.json.JSONString(required=False)


class ContractDetailsUpdateInputType(OpenIMISMutation.Input):
    id = graphene.UUID(required=True)
    contract_id = graphene.UUID(required=False)
    insuree_id = graphene.Int(required=False)
    contribution_plan_bundle_id = graphene.UUID(required=False)
    json_ext = graphene.types.json.JSONString(required=False)
    json_param = graphene.types.json.JSONString(required=False)


class ContractDetailsCreateFromInsureeInputType(OpenIMISMutation.Input):
    contract_id = graphene.UUID(required=True)
    policy_holder_insuree_id = graphene.UUID(required=True)


# Generated by Django 3.0.3 on 2020-12-28 12:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contract', '0005_auto_20201228_1151'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='contractcontributionplandetails',
            name='date_valid_from',
        ),
        migrations.RemoveField(
            model_name='contractcontributionplandetails',
            name='date_valid_to',
        ),
        migrations.RemoveField(
            model_name='contractcontributionplandetails',
            name='replacement_uuid',
        ),
        migrations.RemoveField(
            model_name='historicalcontractcontributionplandetails',
            name='date_valid_from',
        ),
        migrations.RemoveField(
            model_name='historicalcontractcontributionplandetails',
            name='date_valid_to',
        ),
        migrations.RemoveField(
            model_name='historicalcontractcontributionplandetails',
            name='replacement_uuid',
        ),
    ]



# Generated by Django 3.0.3 on 2021-02-02 08:15

import core.fields
import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contract', '0012_contractdetailsmutation'),
    ]

    operations = [
        migrations.AddField(
            model_name='contractcontributionplandetails',
            name='date_valid_from',
            field=core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now),
        ),
        migrations.AddField(
            model_name='contractcontributionplandetails',
            name='date_valid_to',
            field=core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True),
        ),
        migrations.AddField(
            model_name='contractcontributionplandetails',
            name='replacement_uuid',
            field=models.UUIDField(db_column='ReplacementUUID', null=True),
        ),
        migrations.AddField(
            model_name='historicalcontractcontributionplandetails',
            name='date_valid_from',
            field=core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now),
        ),
        migrations.AddField(
            model_name='historicalcontractcontributionplandetails',
            name='date_valid_to',
            field=core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True),
        ),
        migrations.AddField(
            model_name='historicalcontractcontributionplandetails',
            name='replacement_uuid',
            field=models.UUIDField(db_column='ReplacementUUID', null=True),
        ),
    ]



import datetime
import numbers
import base64
from unittest import mock
from django.test import TestCase
from uuid import UUID

import graphene
from contract.tests.helpers import *
from contract import schema as contract_schema
from graphene import Schema
from graphene.test import Client


class ContractQueryTest(TestCase):

    class BaseTestContext:
        user = mock.Mock(is_anonymous=False)
        user.has_perm = mock.MagicMock(return_value=False)

    class AnonymousUserContext:
        user = mock.Mock(is_anonymous=True)

    @classmethod
    def setUpClass(cls):
        cls.date_created = datetime.datetime.now()
        cls.test_contract = create_test_contract(
            custom_props={
                'code': 'testContract-'+str(cls.date_created),
                'payment_reference': 'payment reference'+str(cls.date_created),
                'amount_due': 450.99,
                'date_valid_from':  datetime.date(2020, 1, 1),
                'amendment': 1
            }
        )
        cls.test_contract_details = create_test_contract_details(
            contract=cls.test_contract
        )
        cls.test_contract_contribution_plan_details = create_test_contract_contribution_plan_details(
            contract_details=cls.test_contract_details
        )

        cls.schema = Schema(
            query=contract_schema.Query,
        )

        cls.graph_client = Client(cls.schema)

    @classmethod
    def tearDownClass(cls):
        ContractContributionPlanDetails.objects.filter(id=cls.test_contract_contribution_plan_details.id).delete()
        ContractDetails.objects.filter(id=cls.test_contract_details.id).delete()
        Contract.objects.filter(id=cls.test_contract.id).delete()

    def test_find_contract_existing(self):
        id = self.test_contract.id
        result = self.find_by_id_query("contract", id)
        converted_id = base64.b64decode(result[0]['node']['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_details_existing(self):
        id = self.test_contract_details.id
        result = self.find_by_id_query("contractDetails", id)
        converted_id = base64.b64decode(result[0]['node']['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_existing(self):
        id = self.test_contract_contribution_plan_details.id
        result = self.find_by_id_query("contractContributionPlanDetails", id)
        converted_id = base64.b64decode(result[0]['node']['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_params(self):
        expected = self.test_contract
        params = {
            'version': expected.version,
            'isDeleted': True if expected.is_deleted else False,
            'code': expected.code,
        }
        result = self.find_by_exact_attributes_query("contract", params)
        self.assertDictEqual(result[0]['node'], params)

    def test_find_contract_details_by_contract(self):
        details_contract_id = self.test_contract_details.contract.id
        id = self.test_contract_details.id
        query = F'''
        {{
            contractDetails(
                contract_Id: "{details_contract_id}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractDetails']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_details_by_id_and_insuree(self):
        details_insuree_uuid = self.test_contract_details.insuree.uuid
        id = self.test_contract_details.id
        query = F'''
        {{
            contractDetails(
                id: "{id}"
                insuree_Uuid: "{details_insuree_uuid}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractDetails']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code(self):
        code = self.test_contract.code
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                code: "{code}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_payment_reference(self):
        payment_reference = self.test_contract.payment_reference
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                paymentReference: "{payment_reference}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_amount_due(self):
        code = self.test_contract.code
        amount_due = self.test_contract.amount_due
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                code:" {code}", amountDue: {amount_due}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_amount_due_greater(self):
        code = self.test_contract.code
        amount_due = self.test_contract.amount_due - 100
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                code:" {code}", amountDue_Gte: {amount_due}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_amount_due_greater_equal(self):
        code = self.test_contract.code
        amount_due = self.test_contract.amount_due
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                code:" {code}", amountDue_Gte: {amount_due}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_amound_due_lt(self):
        code = self.test_contract.code
        amount_due = self.test_contract.amount_due
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                code:" {code}", amountDue_Lt: {amount_due}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges']
        self.assertEqual(len(result), 0)

    def test_find_contract_by_insuree(self):
        details_insuree_uuid = self.test_contract_details.insuree.uuid
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                insuree: "{details_insuree_uuid}") {{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_id_and_state(self):
        details_contract_state = self.test_contract.state
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                id: "{id}"
                state: {details_contract_state}) {{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_contract_details_and_contribution_plan(self):
        details_contract_details_id = self.test_contract_contribution_plan_details.contract_details.id
        details_contribution_plan_id = self.test_contract_contribution_plan_details.contribution_plan.id
        id = self.test_contract_contribution_plan_details.id
        query = F'''
        {{
            contractContributionPlanDetails(
                contractDetails_Id: "{details_contract_details_id}",
                contributionPlan_Id: "{details_contribution_plan_id}") {{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractContributionPlanDetails']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_date_valid_from_gte(self):
        code = self.test_contract.code
        date_valid_from = str(self.test_contract.date_valid_from)+'T00:00:00'
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                code: "{code}", dateValidFrom_Gte: "{date_valid_from}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_contains(self):
        date = self.date_created
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                code_Icontains: "{date}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_code_amendment(self):
        code = self.test_contract.code
        amendment = self.test_contract.amendment
        id = self.test_contract.id
        query = F'''
        {{
            contract(
                code: "{code}", amendment: {amendment}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_start_with(self):
        query = F'''
        {{
            contract(
                code_Istartswith: "testContract-"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contract']['edges']
        self.assertGreater(len(result), 0)

    def test_find_contracts_details_by_date_created(self):
        date_created = str(self.test_contract_details.date_created).replace(' ', 'T')
        id = self.test_contract_details.id
        query = F'''
        {{
            contractDetails(
                dateCreated: "{date_created}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractDetails']['edges'][0]["node"]
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contracts_details_by_date_updated(self):
        date_updated = str(self.test_contract_details.date_updated).replace(' ', 'T')
        id = self.test_contract_details.id
        query = F'''
        {{
            contractDetails(
                dateUpdated: "{date_updated}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractDetails']['edges'][0]["node"]
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_date_created(self):
        date_created = str(self.test_contract_contribution_plan_details.date_created).replace(' ', 'T')
        id = self.test_contract_contribution_plan_details.id
        query = F'''
        {{
            contractContributionPlanDetails(
                dateCreated: "{date_created}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractContributionPlanDetails']['edges'][0]["node"]
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_date_created_gte(self):
        date_created = str(self.test_contract_contribution_plan_details.date_created).replace(' ', 'T')
        id = self.test_contract_contribution_plan_details.id
        query = F'''
        {{
            contractContributionPlanDetails(
                dateCreated_Gte: "{date_created}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractContributionPlanDetails']['edges'][0]["node"]
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_date_created_gt(self):
        date_created = str(self.test_contract_contribution_plan_details.date_created).replace(' ', 'T')
        id = self.test_contract_contribution_plan_details.id
        query = F'''
        {{
            contractContributionPlanDetails(
                dateCreated_Gt: "{date_created}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractContributionPlanDetails']['edges']
        self.assertEqual(len(result), 0)

    def test_find_contract_contribution_plan_details_by_date_updated(self):
        date_updated = str(self.test_contract_contribution_plan_details.date_updated).replace(' ', 'T')
        id = self.test_contract_contribution_plan_details.id
        query = F'''
        {{
            contractContributionPlanDetails(
                dateUpdated: "{date_updated}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractContributionPlanDetails']['edges'][0]["node"]
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_id_and_contribution(self):
        details_contribution_uuid = self.test_contract_contribution_plan_details.contribution.uuid
        id = self.test_contract_contribution_plan_details.id
        query = F'''
        {{
            contractContributionPlanDetails(
                id: "{id}"
                contribution_Uuid: "{details_contribution_uuid}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractContributionPlanDetails']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_id_and_amount_in_contribution_gte(self):
        details_contribution_amount = self.test_contract_contribution_plan_details.contribution.amount
        id = self.test_contract_contribution_plan_details.id
        query = F'''
        {{
            contractContributionPlanDetails(
                id: "{id}"
                contribution_Amount_Gte: {details_contribution_amount}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractContributionPlanDetails']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_id_and_amount_in_contribution_gt(self):
        details_contribution_amount = self.test_contract_contribution_plan_details.contribution.amount
        id = self.test_contract_contribution_plan_details.id
        query = F'''
        {{
            contractContributionPlanDetails(
                id: "{id}"
                contribution_Amount_Gt: {details_contribution_amount}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractContributionPlanDetails']['edges']
        self.assertEqual(len(result), 0)

    def test_find_contract_contribution_plan_details_by_insuree(self):
        details_insuree_uuid = self.test_contract_details.insuree.uuid
        id = self.test_contract_contribution_plan_details.id
        query = F'''
        {{
            contractContributionPlanDetails(
                insuree: "{details_insuree_uuid}") {{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query)
        result = query_result['contractContributionPlanDetails']['edges'][0]['node']
        converted_id = base64.b64decode(result['id']).decode('utf-8').split(':')[1]
        self.assertEqual(UUID(converted_id), id)

    def find_by_id_query(self, query_type, id, context=None):
        query = F'''
        {{
            {query_type}(id:"{id}") {{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        '''

        query_result = self.execute_query(query, context=context)
        records = query_result[query_type]['edges']

        if len(records) > 1:
            raise ValueError(F"Ambiguous id {id} for query {query_type}")

        return records

    def find_by_exact_attributes_query(self, query_type, params, context=None):
        node_content_str = "\n".join(params.keys())
        query = F'''
        {{
            {query_type}({self.build_params(params)}) {{
                totalCount
                edges {{
                  node {{
                    {node_content_str}
                  }}
                  cursor
                }}
          }}
        }}
        '''
        query_result = self.execute_query(query, context=context)
        records = query_result[query_type]['edges']
        return records

    def execute_query(self, query, context=None):
        if context is None:
            context = self.BaseTestContext()

        query_result = self.graph_client.execute(query, context=context)
        query_data = query_result['data']
        return query_data

    def build_params(self, params):
        def wrap_arg(v):
            if isinstance(v, str):
                return F'"{v}"'
            if isinstance(v, bool):
                return str(v).lower()
            if isinstance(v, datetime.date):
                return graphene.DateTime.serialize(
                    datetime.datetime.fromordinal(v.toordinal()))
            return v

        params_as_args = [f'{k}:{wrap_arg(v)}' for k, v in params.items()]
        return ", ".join(params_as_args)


from django.apps import AppConfig


MODULE_NAME = "contract"


DEFAULT_CFG = {
    "gql_query_contract_perms": ["152101"],
    "gql_query_contract_admins_perms": [],
    "gql_mutation_create_contract_perms": ["152102"],
    "gql_mutation_update_contract_perms": ["152103"],
    "gql_mutation_delete_contract_perms": ["152104"],
    "gql_mutation_renew_contract_perms": ["152106"],
    "gql_mutation_submit_contract_perms": ["152107"],
    "gql_mutation_approve_ask_for_change_contract_perms": ["152108"],
    "gql_mutation_amend_contract_perms": ["152109"],
    "gql_query_payment_perms": ["101401"],
    "gql_mutation_create_payments_perms": ["101402"],
    "gql_mutation_update_payments_perms": ["101403"],
    "gql_mutation_delete_payments_perms": ["101404"],
    "gql_mutation_approve_payments_perms": ["101408"],
    # OFS-259: Support the policyholder portal perms on Contract
    "gql_query_contract_policyholder_portal_perms": ["154201"],
    "gql_mutation_create_contract_policyholder_portal_perms": ["154202"],
    "gql_mutation_update_contract_policyholder_portal_perms": ["154203"],
    "gql_mutation_submit_contract_policyholder_portal_perms": ["154207"],
    "gql_mutation_amend_contract_policyholder_portal_perms": ["154209"],
    "gql_invoice_create_perms": ["155102"],
}


class ContractConfig(AppConfig):
    name = MODULE_NAME
    gql_query_contract_perms = []
    gql_query_contract_admins_perms = []
    gql_mutation_create_contract_perms = []
    gql_mutation_update_contract_perms = []
    gql_mutation_delete_contract_perms = []
    gql_mutation_renew_contract_perms = []
    gql_mutation_submit_contract_perms = []
    gql_mutation_approve_ask_for_change_contract_perms = []
    gql_mutation_amend_contract_perms = []
    gql_mutation_create_payments_perms = []
    gql_mutation_update_payments_perms = []
    gql_mutation_delete_payments_perms = []
    gql_mutation_approve_payments_perms = []
    # OFS-259: Support the policyholder portal perms on Contract
    gql_query_contract_policyholder_portal_perms = []
    gql_mutation_create_contract_policyholder_portal_perms = []
    gql_mutation_update_contract_policyholder_portal_perms = []
    gql_mutation_submit_contract_policyholder_portal_perms = []
    gql_mutation_amend_contract_policyholder_portal_perms = []
    gql_invoice_create_perms = []

    def _configure_permissions(selfself, cfg):
        ContractConfig.gql_query_contract_perms = cfg[
            "gql_query_contract_perms"]
        ContractConfig.gql_query_contract_admins_perms = cfg[
            "gql_query_contract_admins_perms"
        ]
        ContractConfig.gql_mutation_create_contract_perms = cfg[
            "gql_mutation_create_contract_perms"]
        ContractConfig.gql_mutation_update_contract_perms = cfg[
            "gql_mutation_update_contract_perms"
        ]
        ContractConfig.gql_mutation_delete_contract_perms = cfg[
            "gql_mutation_delete_contract_perms"]
        ContractConfig.gql_mutation_renew_contract_perms = cfg[
            "gql_mutation_renew_contract_perms"
        ]
        ContractConfig.gql_mutation_submit_contract_perms = cfg[
            "gql_mutation_submit_contract_perms"
        ]
        ContractConfig.gql_mutation_approve_ask_for_change_contract_perms = cfg[
            "gql_mutation_approve_ask_for_change_contract_perms"
        ]
        ContractConfig.gql_mutation_amend_contract_perms = cfg[
            "gql_mutation_amend_contract_perms"
        ]
        ContractConfig.gql_mutation_create_payments_perms = cfg[
            "gql_mutation_create_payments_perms"
        ]
        ContractConfig.gql_mutation_update_payments_perms = cfg[
            "gql_mutation_update_payments_perms"
        ]
        ContractConfig.gql_mutation_delete_payments_perms = cfg[
            "gql_mutation_delete_payments_perms"
        ]
        ContractConfig.gql_mutation_approve_payments_perms = cfg[
            "gql_mutation_approve_payments_perms"
        ]
        # OFS-259: Support the policyholder portal perms on Contract
        ContractConfig.gql_query_contract_policyholder_portal_perms = cfg[
            "gql_query_contract_policyholder_portal_perms"
        ]
        ContractConfig.gql_mutation_create_contract_policyholder_portal_perms = cfg[
            "gql_mutation_create_contract_policyholder_portal_perms"
        ]
        ContractConfig.gql_mutation_update_contract_policyholder_portal_perms = cfg[
            "gql_mutation_update_contract_policyholder_portal_perms"
        ]
        ContractConfig.gql_mutation_submit_contract_policyholder_portal_perms = cfg[
            "gql_mutation_submit_contract_policyholder_portal_perms"
        ]
        ContractConfig.gql_mutation_amend_contract_policyholder_portal_perms = cfg[
            "gql_mutation_amend_contract_policyholder_portal_perms"
        ]
        ContractConfig.gql_invoice_create_perms = cfg[
            "gql_invoice_create_perms"
        ]

    def ready(self):
        from core.models import ModuleConfiguration
        cfg = ModuleConfiguration.get_or_default(MODULE_NAME, DEFAULT_CFG)
        self._configure_permissions(cfg)
        import contract.signals


import logging

from celery import shared_task
from core.models import User
from contract.models import Contract, ContractContributionPlanDetails
from contract.services import Contract as ContractService, ContractToInvoiceService


logger = logging.getLogger(__name__)


@shared_task
def approve_contracts(user_id, contracts):
    user = User.objects.get(id=user_id)
    contract_service = ContractService(user=user)
    for contract in contracts:
        output = contract_service.approve(contract={"id": contract})


@shared_task
def counter_contracts(user_id, contracts):
    user = User.objects.get(id=user_id)
    contract_service = ContractService(user=user)
    for contract in contracts:
        output = contract_service.counter(contract={"id": contract})


@shared_task
def create_invoice_from_contracts(user_id, contracts):
    user = User.objects.get(id=user_id)
    contract_service = ContractToInvoiceService(user=user)
    for contract in contracts:
        contract_instance = Contract.objects.filter(id=contract)
        if contract_instance:
            contract_instance = contract_instance.first()
            ccpd_list = ContractContributionPlanDetails.objects.filter(contract_details__contract=contract_instance)
            output = contract_service.create_invoice(
                instance=contract_instance,
                convert_to="InvoiceLine",
                user=user,
                ccpd_list=ccpd_list
            )



import logging

from django.db import migrations

logger = logging.getLogger(__name__)


MIGRATION_SQL = """
    /* Contract */
    DECLARE @SystemRole INT
    SELECT @SystemRole = role.RoleID from tblRole role where IsSystem=256;
    /* Contract aprrove or counter */
    IF NOT EXISTS (SELECT * FROM [tblRoleRight] WHERE [RoleID] = @SystemRole AND [RightID] = 152109)
    BEGIN
        INSERT [dbo].[tblRoleRight] ([RoleID], [RightID], [ValidityFrom]) 
        VALUES (@SystemRole, 152109, CURRENT_TIMESTAMP)
    END 
"""


class Migration(migrations.Migration):
    dependencies = [
        ('contract', '0018_approve_ask_for_change_perms')
    ]

    operations = [
        migrations.RunSQL(MIGRATION_SQL)
    ]



from django.test import TestCase
from contract.services import Contract as ContractService, ContractDetails as ContractDetailsService, \
    ContractContributionPlanDetails as ContractContributionPlanDetailsService
from contract.models import Contract, ContractDetails, ContractContributionPlanDetails
from core.test_helpers import create_test_technical_user
from policyholder.models import PolicyHolder, PolicyHolderInsuree
from policyholder.tests.helpers import create_test_policy_holder, create_test_policy_holder_insuree
from contribution_plan.tests.helpers import create_test_contribution_plan, \
    create_test_contribution_plan_bundle, create_test_contribution_plan_bundle_details
from contribution_plan.models import ContributionPlan, ContributionPlanBundle, ContributionPlanBundleDetails
from policy.test_helpers import create_test_policy
from core.models import User
from calculation.services import get_parameters, get_rule_details, \
    run_calculation_rules, get_rule_name, get_linked_class


class ServiceTestContract(TestCase):
    user = None

    @classmethod
    def setUpClass(cls):
        cls.user = User.objects.filter(username='admin').first()
        if not cls.user:
            cls.user = create_test_technical_user(username='admin', password='S\/pe®Pąßw0rd™', super_user=True)
        cls.contract_service = ContractService(cls.user)
        cls.contract_details_service = ContractDetailsService(cls.user)
        cls.contract_contribution_plan_details_service = ContractContributionPlanDetailsService(cls.user)
        # some test data so as to created contract properly
        cls.income = 500
        cls.rate = 5
        cls.number_of_insuree = 5
        cls.policy_holder = create_test_policy_holder()
        cls.policy_holder2 = create_test_policy_holder()
        # create contribution plans etc
        cls.contribution_plan_bundle = create_test_contribution_plan_bundle()
        cls.contribution_plan = create_test_contribution_plan(
            custom_props={"json_ext": '{"calculation_rule":{"rate": "' + str(cls.rate) + '"}}'}
        )
        cls.contribution_plan_bundle_details = create_test_contribution_plan_bundle_details(
            contribution_plan=cls.contribution_plan,
            contribution_plan_bundle=cls.contribution_plan_bundle
        )
        from core import datetime
        # create policy holder insuree for that test policy holder
        for i in range(0, cls.number_of_insuree):
            ph_insuree = create_test_policy_holder_insuree(
                policy_holder=cls.policy_holder,
                contribution_plan_bundle=cls.contribution_plan_bundle,
                custom_props={
                    "last_policy": None,
                    "json_ext": '{"calculation_rule":{"income": "' + str(cls.income) + '"}}'
                }
            )
            create_test_policy(
                cls.contribution_plan.benefit_plan,
                ph_insuree.insuree,
                custom_props={
                    "start_date": datetime.datetime(2016, 3, 1),
                    "expiry_date": datetime.datetime(2021, 7, 1)
                }
            )

    @classmethod
    def tearDownClass(cls):
        # tear down the test data created during set up
        PolicyHolderInsuree.objects.filter(policy_holder_id=str(cls.policy_holder.id)).delete()
        PolicyHolder.objects.filter(id=cls.policy_holder.id).delete()
        ContributionPlanBundleDetails.objects.filter(id=cls.contribution_plan_bundle_details.id).delete()
        ContributionPlanBundle.objects.filter(id=cls.contribution_plan_bundle.id).delete()
        ContributionPlan.objects.filter(id=cls.contribution_plan.id).delete()

    def test_contract_create_without_policy_holder(self):
        contract = {
            'code': 'AAAAAA',
        }
        response = self.contract_service.create(contract)
        # tear down the test data
        Contract.objects.filter(id=response["data"]["id"]).delete()
        self.assertEqual(
            (
                 True,
                 "Ok",
                 "",
                 "AAAAAA",
                 0,
                 None,
            ),
            (
                 response['success'],
                 response['message'],
                 response['detail'],
                 response['data']['code'],
                 response['data']['amendment'],
                 response['data']['amount_notified'],
            )
        )

    def test_contract_create_with_policy_holder(self):
        contract = {
            'code': 'TESTONE',
            'policy_holder_id': self.policy_holder.id
        }
        response = self.contract_service.create(contract)
        # tear down the test data
        ContractDetails.objects.filter(contract_id=response["data"]["id"]).delete()
        Contract.objects.filter(id=response["data"]["id"]).delete()
        self.assertEqual(
            (
                 True,
                 "Ok",
                 "",
                 "TESTONE",
                 0,
            ),
            (
                 response['success'],
                 response['message'],
                 response['detail'],
                 response['data']['code'],
                 response['data']['amendment'],
            )
        )

    def test_contract_create_update_delete_with_policy_holder(self):
        contract = {
            "code": "CTSV",
            "policy_holder_id": str(self.policy_holder.id)
        }
        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])

        contract = {
            "id": contract_id,
            "payment_reference": "payment_one xxxxxxxx",
        }
        response = self.contract_service.update(contract)
        updated_payment_reference = response['data']['payment_reference']

        contract = {
            "id": contract_id,
        }
        response = self.contract_service.delete(contract)
        is_deleted = response['success']

        # tear down the test data
        ContractDetails.objects.filter(contract_id=contract_id).delete()
        Contract.objects.filter(id=contract_id).delete()

        self.assertEqual(
            ("payment_one xxxxxxxx", True),
            (updated_payment_reference, is_deleted)
        )

    def test_contract_create_update_failed_ph(self):
        contract = {
            "code": "CSTG",
            "policy_holder_id": str(self.policy_holder.id)
        }
        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])

        contract = {
            "id": contract_id,
            "policy_holder_id": str(self.policy_holder2.id),
        }
        response = self.contract_service.update(contract)
        failed = response['detail']

        # tear down the test data
        ContractDetails.objects.filter(contract_id=contract_id).delete()
        Contract.objects.filter(id=contract_id).delete()

        self.assertEqual(
            "ContractUpdateError: You cannot update already set PolicyHolder in Contract!", failed,
        )

    def test_contract_create_submit_fail_scenarios(self):
        contract = {
            "code": "MTD",
            "policy_holder_id": str(self.policy_holder.id)
        }
        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])

        contract_created = Contract.objects.filter(id=contract_id).first()
        contract_created.state = Contract.STATE_EXECUTABLE
        contract_created.save(username="admin")

        contract = {
            "id": contract_id,
        }

        response = self.contract_service.submit(contract)
        result_message = response["detail"]
        expected_message = "ContractUpdateError: The contract cannot be submitted because of current state!"

        contract_created.state = Contract.STATE_NEGOTIABLE
        contract_created.save(username="admin")

        response = self.contract_service.submit(contract)
        result_message2 = response["detail"]
        expected_message2 = "ContractUpdateError: The contract has been already submitted!"

        contract_created.policy_holder = None
        contract_created.save(username="admin")

        response = self.contract_service.submit(contract)
        result_message3 = response["detail"]
        expected_message3 = "ContractUpdateError: The contract does not contain PolicyHolder!"

        # tear down the test data
        list_cd = list(ContractDetails.objects.filter(contract_id=contract_id).values('id', 'json_ext'))
        ContractDetails.objects.filter(contract_id=contract_id).delete()
        Contract.objects.filter(id=contract_id).delete()

        self.assertEqual(
            (
                expected_message,
                expected_message2,
                expected_message3
            ),
            (
                result_message,
                result_message2,
                result_message3
            )
        )

    def test_contract_create_submit_counter(self):
        contract = {
            "code": "SUR",
            "policy_holder_id": str(self.policy_holder.id)
        }
        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])

        contract = {
            "id": contract_id,
        }
        self.contract_service.submit(contract)
        response = self.contract_service.counter(contract)
        result_state = response["data"]["state"]
        expected_state = 11

        # tear down the test data
        list_cd = list(ContractDetails.objects.filter(contract_id=contract_id).values('id'))
        for cd in list_cd:
            ccpd = ContractContributionPlanDetails.objects.filter(contract_details__id=f"{cd['id']}").delete()
        ContractDetails.objects.filter(contract_id=contract_id).delete()
        Contract.objects.filter(id=contract_id).delete()

        self.assertEqual(
            expected_state, result_state
        )

    def test_contract_create_submit(self):
        from core import datetime
        contract = {
            "code": "TESTCON",
            "policy_holder_id": str(self.policy_holder.id),
            "date_valid_from": datetime.datetime(2021, 1, 1),
            "date_valid_to": datetime.datetime(2023, 6, 30),
        }
        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])
        contract = {"id": contract_id,}
        response = self.contract_service.submit(contract)
        expected_state = 4
        result_state = response["data"]["state"]
        # tear down the test data
        list_cd = list(ContractDetails.objects.filter(contract_id=contract_id).values('id'))
        for cd in list_cd:
            ContractContributionPlanDetails.objects.filter(contract_details__id=f"{cd['id']}").delete()
        ContractDetails.objects.filter(contract_id=contract_id).delete()
        Contract.objects.filter(id=contract_id).delete()
        self.assertEqual(
            expected_state, result_state
        )

    def test_contract_create_cd_from_phi(self):
        from core import datetime
        ph_insuree2 = create_test_policy_holder_insuree(
            policy_holder=self.policy_holder,
            contribution_plan_bundle=self.contribution_plan_bundle,
            custom_props={"last_policy": None, "json_ext": '{"calculation_rule":{"income": "400"}}'}
        )
        contract = {
            "code": "MTEST-1",
            "policy_holder_id": str(self.policy_holder.id),
            "date_valid_from": datetime.datetime(2021, 1, 1),
            "date_valid_to": datetime.datetime(2023, 6, 30),
        }
        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])
        contract = {"id": contract_id,}
        ph_insuree_input = {"id": f'{ph_insuree2.id}',}
        response = self.contract_details_service.ph_insuree_to_contract_details(
            contract=contract,
            ph_insuree=ph_insuree_input
        )
        # tear down the test data
        ContractDetails.objects.filter(contract_id=contract_id).delete()
        Contract.objects.filter(id=contract_id).delete()

        self.assertEqual(
            True, response["success"]
        )


class CalculationContractTest(TestCase):
    user = None

    @classmethod
    def setUpClass(cls):
        cls.user = User.objects.filter(username='admin').first()
        if not cls.user:
            cls.user = create_test_technical_user(username='admin', password='S\/pe®Pąßw0rd™', super_user=True)
        cls.contract_service = ContractService(cls.user)
        cls.income = 500
        cls.rate = 5
        cls.number_of_insuree = 5
        cls.policy_holder = create_test_policy_holder()
        # create contribution plans etc
        cls.contribution_plan_bundle = create_test_contribution_plan_bundle()
        cls.contribution_plan = create_test_contribution_plan(
            custom_props={"json_ext": '{"calculation_rule":{"rate": "' + str(cls.rate) + '"}}'}
        )
        cls.contribution_plan_bundle_details = create_test_contribution_plan_bundle_details(
            contribution_plan=cls.contribution_plan,
            contribution_plan_bundle=cls.contribution_plan_bundle
        )

        # create policy holder insuree for that test policy holder
        for i in range(0, cls.number_of_insuree):
            create_test_policy_holder_insuree(
                policy_holder=cls.policy_holder,
                contribution_plan_bundle=cls.contribution_plan_bundle,
                custom_props={"last_policy": None, "json_ext": '{"calculation_rule":{"income": "' + str(cls.income) + '"}}'}
            )

    @classmethod
    def tearDownClass(cls):
        # tear down the test data created during set up
        PolicyHolderInsuree.objects.filter(policy_holder_id=str(cls.policy_holder.id)).delete()
        PolicyHolder.objects.filter(id=cls.policy_holder.id).delete()
        ContributionPlanBundleDetails.objects.filter(id=cls.contribution_plan_bundle_details.id).delete()
        ContributionPlanBundle.objects.filter(id=cls.contribution_plan_bundle.id).delete()
        ContributionPlan.objects.filter(id=cls.contribution_plan.id).delete()

    def test_get_rule_name(self):
        class_name = "ContractDetails"
        result = get_rule_name(class_name=class_name)
        self.assertEqual("ContributionValuationRule", result[0][1].__name__)

    def test_get_rule_not_existing(self):
        class_name = "xxxxxxxxxxxxxxxxxxx"
        result = get_rule_name(class_name=class_name)
        self.assertEqual(None, result[0][1])

    def test_get_rule_details(self):
        class_name = "PolicyHolderInsuree"
        class_name2 = "ContributionPlan"
        result = get_rule_details(class_name=class_name)[0][1]
        result2 = get_rule_details(class_name=class_name2)[0][1]
        result_param = [param["name"] for param in result["parameters"]]
        result2_param = [param["name"] for param in result2["parameters"]]
        self.assertEqual(
            (class_name, class_name2, ["income"], ["rate", "includeFamilly"]),
            (result["class"], result2["class"], result_param, result2_param)
        )

    def test_get_rule_details_not_existing(self):
        class_name = "xxxxxxxxxxxxxxx"
        result = get_rule_details(class_name=class_name)
        self.assertEqual(None, result[0][1])

    def test_get_linked_class_empty(self):
        result = get_linked_class()
        self.assertEqual(['Calculation'], result[0][1])

    def test_get_linked_class(self):
        result = get_linked_class(["PolicyHolderInsuree"])
        self.assertEqual(['PolicyHolder', 'Insuree', 'ContributionPlanBundle', 'Policy'], result[0][1])

    def test_get_param_and_amount_calculation(self):
        # create contract to test 1st calculation rule contribution valuation
        # test case - create contract and check if amount is calcutated properly
        # and test if on instance of contract details the proper params is showed
        # by getting param name
        contract = {
            'code': 'CALTEST',
            'policy_holder_id': self.policy_holder.id
        }

        response = self.contract_service.create(contract)
        # after creating contract - we can get contract details so as to get params
        # run calculation rules etc
        cd = ContractDetails.objects.filter(contract__id=response["data"]["id"]).first()
        result_params = get_parameters("PolicyHolderInsuree", cd)

        # tear down the contract test data
        ContractDetails.objects.filter(contract_id=response["data"]["id"]).delete()
        Contract.objects.filter(id=response["data"]["id"]).delete()

        # we want to assert name of param related to the contract details (should be 'income')
        # and also the value of contract (all contributions)
        # for that test should be income=500, 5 insurees, default rate=5 %
        # income*rate*number of contributions = according to Contribution Valuation Rule
        self.assertEqual(
            ("income", self.income * (float(self.rate / 100)) * self.number_of_insuree),
            (result_params[0][1][0]["name"], response["data"]["amount_notified"])
        )



import uuid

from contribution_plan.models import ContributionPlanBundle, ContributionPlan
from django.conf import settings
from django.db import models
from core import models as core_models, fields
from graphql import ResolveInfo
from jsonfallback.fields import FallbackJSONField
from policy.models import Policy
from contribution.models import Premium
from policyholder.models import PolicyHolder
from insuree.models import Insuree


class ContractManager(models.Manager):
    def filter(self, *args, **kwargs):
        keys = [x for x in kwargs if "itemsvc" in x]
        for key in keys:
            new_key = key.replace("itemsvc", self.model.model_prefix)
            kwargs[new_key] = kwargs.pop(key)
        return super(ContractManager, self).filter(*args, **kwargs)


class Contract(core_models.HistoryBusinessModel):
    code = models.CharField(db_column='Code', max_length=64, null=False)
    policy_holder = models.ForeignKey(PolicyHolder, db_column="PolicyHolderUUID",
                                      on_delete=models.deletion.DO_NOTHING, blank=True, null=True)
    amount_notified = models.FloatField(db_column='AmountNotified', blank=True, null=True)
    amount_rectified = models.FloatField(db_column='AmountRectified', blank=True, null=True)
    amount_due = models.FloatField(db_column='AmountDue', blank=True, null=True)
    date_approved = fields.DateTimeField(db_column='DateApproved', blank=True, null=True)
    date_payment_due = fields.DateField(db_column='DatePaymentDue', blank=True, null=True)
    state = models.SmallIntegerField(db_column='State', blank=True, null=True)
    payment_reference = models.CharField(db_column='PaymentReference', max_length=255, blank=True, null=True)
    amendment = models.IntegerField(db_column='Amendment', blank=False, null=False, default=0)

    objects = ContractManager()

    @property
    def amount(self):
        amount = 0
        if self.state in [1, 2]:
            amount = self.amount_notified
        elif self.state in [4, 11, 3]:
            amount = self.amount_rectified
        elif self.state in [5, 6, 7, 8, 9, 10]:
            amount = self.amount_due
        else:
            amount = self.amount_due
        return amount

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        db_table = 'tblContract'

    STATE_REQUEST_FOR_INFORMATION = 1
    STATE_DRAFT = 2
    STATE_OFFER = 3
    STATE_NEGOTIABLE = 4
    STATE_EXECUTABLE = 5
    STATE_ADDENDUM = 6
    STATE_EFFECTIVE = 7
    STATE_EXECUTED = 8
    STATE_DISPUTED = 9
    STATE_TERMINATED = 10
    STATE_COUNTER = 11


class ContractDetailsManager(models.Manager):
    def filter(self, *args, **kwargs):
        keys = [x for x in kwargs if "itemsvc" in x]
        for key in keys:
            new_key = key.replace("itemsvc", self.model.model_prefix)
            kwargs[new_key] = kwargs.pop(key)
        return super(ContractDetailsManager, self).filter(*args, **kwargs)


class ContractDetails(core_models.HistoryModel):
    contract = models.ForeignKey(Contract, db_column="ContractUUID",
                                      on_delete=models.deletion.CASCADE)
    insuree = models.ForeignKey(Insuree, db_column='InsureeID',
                                              on_delete=models.deletion.DO_NOTHING)
    contribution_plan_bundle = models.ForeignKey(ContributionPlanBundle,
                                                 db_column='ContributionPlanBundleUUID',
                                                 on_delete=models.deletion.DO_NOTHING)

    json_param = FallbackJSONField(db_column='Json_param', blank=True, null=True)

    objects = ContractDetailsManager()

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        db_table = 'tblContractDetails'


class ContractContributionPlanDetailsManager(models.Manager):
    def filter(self, *args, **kwargs):
        keys = [x for x in kwargs if "itemsvc" in x]
        for key in keys:
            new_key = key.replace("itemsvc", self.model.model_prefix)
            kwargs[new_key] = kwargs.pop(key)
        return super(ContractContributionPlanDetailsManager, self).filter(*args, **kwargs)


class ContractContributionPlanDetails(core_models.HistoryBusinessModel):
    contribution_plan = models.ForeignKey(ContributionPlan, db_column='ContributionPlanUUID',
                                          on_delete=models.deletion.DO_NOTHING)
    policy = models.ForeignKey(Policy, db_column='PolicyID',
                               on_delete=models.deletion.DO_NOTHING)
    contract_details = models.ForeignKey(ContractDetails, db_column='ContractDetailsUUID',
                                         on_delete=models.deletion.CASCADE)
    contribution = models.ForeignKey(Premium, db_column='ContributionId',
                                     related_name='contract_contribution_plan_details', on_delete=models.deletion.DO_NOTHING,
                                     blank=True, null=True)

    objects = ContractContributionPlanDetailsManager()

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        db_table = 'tblContractContributionPlanDetails'


class ContractMutation(core_models.UUIDModel, core_models.ObjectMutation):
    contract = models.ForeignKey(Contract, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(core_models.MutationLog, models.DO_NOTHING, related_name='contracts')

    class Meta:
        managed = True
        db_table = "contract_contractMutation"


class ContractDetailsMutation(core_models.UUIDModel, core_models.ObjectMutation):
    contract_detail = models.ForeignKey(ContractDetails, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(core_models.MutationLog, models.DO_NOTHING, related_name='contract_details')

    class Meta:
        managed = True
        db_table = "contract_contractDetailsMutation"


import json
import datetime

from contribution_plan.tests.helpers import create_test_contribution_plan_bundle, create_test_contribution_plan
from core.models import InteractiveUser, User
from insuree.test_helpers import create_test_insuree
from policy.test_helpers import create_test_policy
from product.test_helpers import create_test_product
from policyholder.tests.helpers import create_test_policy_holder
from contribution.models import Payer, Premium
from contract.models import Contract, ContractDetails, ContractContributionPlanDetails


def create_test_contract(policy_holder=None, custom_props={}):
    if not policy_holder:
        policy_holder = create_test_policy_holder()

    user = __get_or_create_simple_contract_user()

    object_data = {
            'code': "CON",
            'policy_holder': policy_holder,
            'amount_notified': 0,
            'amount_rectified': 0,
            'amount_due': 0,
            'date_payment_due': datetime.date(2011, 10, 31),
            'state': 1,
            'payment_reference': "Payment Reference",
            'json_ext': json.dumps("{}"),
            **custom_props
    }

    contract = Contract(**object_data)
    contract.save(username=user.username)

    return contract


def create_test_contract_details(contract=None, insuree=None,
                                 contribution_plan_bundle=None, custom_props={}):
    if not contract:
        contract = create_test_contract()

    if not insuree:
        insuree = create_test_insuree()

    if not contribution_plan_bundle:
        contribution_plan_bundle = create_test_contribution_plan_bundle()

    user = __get_or_create_simple_contract_user()
    object_data = {
        'contract': contract,
        'insuree': insuree,
        'contribution_plan_bundle': contribution_plan_bundle,
        'json_param': json.dumps("{}"),
        **custom_props
    }

    contract_details = ContractDetails(**object_data)
    contract_details.save(username=user.username)

    return contract_details


def create_test_contract_contribution_plan_details(contribution_plan=None, policy=None,
                                                          contract_details=None, contribution=None, custom_props={}):
    if not contribution_plan:
        contribution_plan = create_test_contribution_plan()

    if not policy:
        policy = create_test_policy(
            product=create_test_product("TestCode", custom_props={"insurance_period": 12,}),
            insuree=create_test_insuree())

    if not contract_details:
        contract_details = create_test_contract_details()

    if not contribution:
        contribution = Premium.objects.create(
            **{
                "policy_id": policy.id,
                "payer_id": None,
                "amount": 1000,
                "receipt": "Test receipt",
                "pay_date": "2019-01-01",
                "validity_from": "2019-01-01",
                "audit_user_id": -1,
            }
        )

    user = __get_or_create_simple_contract_user()
    object_data = {
        'contribution_plan': contribution_plan,
        'policy':policy,
        'contract_details': contract_details,
        'contribution': contribution,
        'json_ext': json.dumps("{}"),
        **custom_props
    }

    contract_contribution_plan_details = ContractContributionPlanDetails(**object_data)
    contract_contribution_plan_details.save(username=user.username)

    return contract_contribution_plan_details


def __get_or_create_simple_contract_user():
    if not User.objects.filter(username='admin').exists():
        User.objects.create_superuser(username='admin', password='S\/pe®Pąßw0rd™')
    user = User.objects.filter(username='admin').first()
    return user


import json
import uuid

from .config import get_message_approved_contract
from .models import Contract, ContractDetails, ContractContributionPlanDetails
from core.signals import Signal
from django.db.models import Q
from django.db.models.signals import post_save
from django.core.serializers.json import DjangoJSONEncoder
from django.conf import settings
from django.core.mail import send_mail, BadHeaderError
from django.dispatch import receiver
from insuree.apps import InsureeConfig
from insuree.models import InsureePolicy
from insuree.signals import signal_before_insuree_policy_query
from payment.apps import PaymentConfig
from payment.models import Payment, PaymentDetail
from payment.signals import signal_before_payment_query
from policy.signals import signal_check_formal_sector_for_policy
from policyholder.apps import PolicyholderConfig
from policyholder.models import PolicyHolderUser
from insuree.models import InsureePolicy


_contract_signal_params = ["contract", "user"]
_contract_approve_signal_params = ["contract", "user", "contract_details_list", "service_object", "payment_service", "ccpd_service"]
signal_contract = Signal(providing_args=_contract_signal_params)
signal_contract_approve = Signal(providing_args=_contract_signal_params)


def on_contract_signal(sender, **kwargs):
    contract = kwargs["contract"]
    user = kwargs["user"]
    __save_or_update_contract(contract=contract, user=user)
    return f"contract updated - state {contract.state}"


def on_contract_approve_signal(sender, **kwargs):
    # approve scenario
    user = kwargs["user"]
    contract_to_approve = kwargs["contract"]
    contract_details_list = kwargs["contract_details_list"]
    contract_service = kwargs["service_object"]
    payment_service = kwargs["payment_service"]
    ccpd_service = kwargs["ccpd_service"]
    # contract valuation
    contract_contribution_plan_details = contract_service.evaluate_contract_valuation(
        contract_details_result=contract_details_list,
        save=True
    )
    contract_to_approve.amount_due = contract_contribution_plan_details["total_amount"]
    result = ccpd_service.create_contribution(contract_contribution_plan_details)
    result_payment = __create_payment(contract_to_approve, payment_service, contract_contribution_plan_details)
    # STATE_EXECUTABLE
    from core import datetime
    now = datetime.datetime.now()
    contract_to_approve.date_approved = now
    contract_to_approve.state = 5
    approved_contract = __save_or_update_contract(contract=contract_to_approve, user=user)
    email_contact_name = contract_to_approve.policy_holder.contact_name["contactName"] \
        if "contactName" in contract_to_approve.policy_holder.contact_name \
        else contract_to_approve.policy_holder.contact_name
    email = __send_email_notify_payment(
        code=contract_to_approve.code,
        name=contract_to_approve.policy_holder.trade_name,
        contact_name=email_contact_name,
        amount_due=contract_to_approve.amount_due,
        payment_reference=contract_to_approve.payment_reference,
        email=contract_to_approve.policy_holder.email,
    )
    return approved_contract


# additional filters for payment in 'contract' tab
def append_contract_filter(sender, **kwargs):
    user = kwargs.get("user", None)
    additional_filter = kwargs.get('additional_filter', None)
    if "contract" in additional_filter:
        # then check perms
        if user.has_perms(PaymentConfig.gql_query_payments_perms) or user.has_perms(PolicyholderConfig.gql_query_payment_portal_perms):
            contract_id = additional_filter["contract"]
            contract_to_process = Contract.objects.filter(id=contract_id).first()
            # check if user is linked to ph in policy holder user table
            type_user = f"{user}"
            # related to user object output (i) or (t)
            # check if we have interactive user from current context
            if '(i)' in type_user:
                from core import datetime
                now = datetime.datetime.now()
                ph_user = PolicyHolderUser.objects.filter(
                    Q(policy_holder__id=contract_to_process.policy_holder.id, user__id=user.id)
                ).filter(
                    Q(date_valid_from=None) | Q(date_valid_from__lte=now),
                    Q(date_valid_to=None) | Q(date_valid_to__gte=now)
                ).first()
                if ph_user or user.has_perms(PaymentConfig.gql_query_payments_perms):
                    return Q(
                        payment_details__premium__contract_contribution_plan_details__contract_details__contract__id=contract_id
                    )


# additional filters for InsureePolicy in contract
def append_contract_policy_insuree_filter(sender, **kwargs):
    user = kwargs.get("user", None)
    additional_filter = kwargs.get('additional_filter', None)
    if "contract" in additional_filter:
        # then check perms
        if user.has_perms(InsureeConfig.gql_query_insuree_policy_perms) or user.has_perms(PolicyholderConfig.gql_query_insuree_policy_portal_perms):
            contract_id = additional_filter["contract"]
            contract_to_process = Contract.objects.filter(id=contract_id).first()
            # check if user is linked to ph in policy holder user table
            type_user = f"{user}"
            # related to user object output (i) or (t)
            # check if we have interactive user from current context
            if '(i)' in type_user:
                from core import datetime
                now = datetime.datetime.now()
                ph_user = PolicyHolderUser.objects.filter(
                    Q(policy_holder__id=contract_to_process.policy_holder.id, user__id=user.id)
                ).filter(
                    Q(date_valid_from=None) | Q(date_valid_from__lte=now),
                    Q(date_valid_to=None) | Q(date_valid_to__gte=now)
                ).first()
                if ph_user or user.has_perms(InsureeConfig.gql_query_insuree_policy_perms):
                    policies = list(
                        ContractContributionPlanDetails.objects.filter(contract_details__contract__id=contract_id).values_list("policy", flat=True)
                    )
                    return Q(
                        start_date__gte=contract_to_process.date_valid_from,
                        start_date__lte=contract_to_process.date_valid_to,
                        policy__in=policies
                    )


# check if policy is related to formal sector contract
def formal_sector_policies(sender, **kwargs):
    policy_id = kwargs.get('policy_id', None)
    ccpd = ContractContributionPlanDetails.objects.filter(policy__id=policy_id, is_deleted=False).first()
    if ccpd:
        cd = ccpd.contract_details
        contract = cd.contract
        return contract.policy_holder
    else:
        return None


signal_contract.connect(on_contract_signal, dispatch_uid="on_contract_signal")
signal_contract_approve.connect(on_contract_approve_signal, dispatch_uid="on_contract_approve_signal")
signal_before_payment_query.connect(append_contract_filter)
signal_before_insuree_policy_query.connect(append_contract_policy_insuree_filter)
signal_check_formal_sector_for_policy.connect(formal_sector_policies)


@receiver(post_save, sender=Payment, dispatch_uid="payment_signal_paid")
def activate_contracted_policies(sender, instance, **kwargs):
    received_amount = instance.received_amount if instance.received_amount else 0
    # check if payment is related to the contract
    payment_detail = PaymentDetail.objects.filter(
        payment__id=int(instance.id)
    ).prefetch_related(
        'premium__contract_contribution_plan_details__contract_details__contract'
    ).prefetch_related(
        'premium__contract_contribution_plan_details'
    ).filter(premium__contract_contribution_plan_details__isnull=False)
    if len(list(payment_detail)) > 0:
        if instance.expected_amount <= received_amount:
            contribution_list_id = [pd.premium.id for pd in payment_detail]
            contract_list = Contract.objects.filter(
                contractdetails__contractcontributionplandetails__contribution__id__in=contribution_list_id
            ).distinct()
            ccpd_number = ContractContributionPlanDetails.objects.prefetch_related('contract_details__contract').filter(
                contract_details__contract__in=list(contract_list)
            ).count()
            # 1- check if the contract have payment attached to each contributions
            # (nbr CCPD of all contract in step 0= Paymentdetails in Steps 0)
            if ccpd_number == len(list(payment_detail)):
                for contract in contract_list:
                    if contract.state == Contract.STATE_EXECUTABLE:
                        # get the ccpd related to the currenttly processing contract
                        ccpd_list = list(
                            ContractContributionPlanDetails.objects.prefetch_related(
                                'contract_details__contract').filter(
                                contract_details__contract=contract
                            )
                        )
                        from core import datetime, datetimedelta
                        # TODO support Splitted payment and check that
                        #  the payment match the value of all contributions
                        for ccpd in ccpd_list:
                            insuree = ccpd.contract_details.insuree
                            pi = InsureePolicy.objects.create(
                                **{
                                    "insuree": insuree,
                                    "policy": ccpd.policy,
                                    "enrollment_date": ccpd.date_valid_from,
                                    "start_date": ccpd.date_valid_from,
                                    "effective_date": ccpd.date_valid_from,
                                    "expiry_date": ccpd.date_valid_to + datetimedelta(
                                        ccpd.contribution_plan.get_contribution_length()
                                    ),
                                    "audit_user_id": -1,
                                }
                            )
                        contract.state = Contract.STATE_EFFECTIVE
                        __save_or_update_contract(contract, contract.user_updated)


def __save_json_external(user_id, datetime, message):
    return {
        "comments": [{
            "From": "Portal/webapp",
            "user": user_id,
            "date": datetime,
            "msg": message
        }]
    }


def __save_or_update_contract(contract, user):
    contract.save(username=user.username)
    historical_record = contract.history.all().first()
    contract.json_ext = json.dumps(__save_json_external(
        user_id=historical_record.user_updated.id,
        datetime=historical_record.date_updated,
        message=f"contract updated - state "
                f"{historical_record.state}"
    ), cls=DjangoJSONEncoder)
    contract.save(username=user.username)
    return contract


def __create_payment(contract, payment_service, contract_cpd):
    from core import datetime
    now = datetime.datetime.now()
    # format payment data
    payment_data = {
        "expected_amount": contract.amount_due,
        "request_date": now,
    }
    payment_details_data = payment_service.collect_payment_details(contract_cpd["contribution_plan_details"])
    return payment_service.create(payment=payment_data, payment_details=payment_details_data)


def __send_email_notify_payment(code, name, contact_name, amount_due, payment_reference, email):
    try:
        email = send_mail(
            subject='Contract payment notification',
            message=get_message_approved_contract(
                language=settings.LANGUAGE_CODE.split('-')[0],
                code=code,
                name=name,
                contact_name=contact_name,
                due_amount=amount_due,
                payment_reference=payment_reference
            ),
            from_email= settings.EMAIL_HOST_USER,
            recipient_list=[email],
            fail_silently=False,
        )
        return email
    except BadHeaderError:
        return ValueError('Invalid header found.')


from django.db.models import Q


def filter_amount_contract(arg='amount_from', arg2='amount_to', **kwargs):
    amount_from = kwargs.get(arg)
    amount_to = kwargs.get(arg2)

    status_notified = [1, 2]
    status_rectified = [4, 11, 3]
    status_due = [5, 6, 7, 8, 9, 10]

    # scenario - only amount_to set
    if not amount_from and amount_to:
        return (
             Q(amount_notified__lte=amount_to, state__in=status_notified) |
             Q(amount_rectified__lte=amount_to, state__in=status_rectified) |
             Q(amount_due__lte=amount_to, state__in=status_due)
        )

    # scenario - only amount_from set
    if amount_from and not amount_to:
        return (
            Q(amount_notified__gte=amount_from, state__in=status_notified) |
            Q(amount_rectified__gte=amount_from, state__in=status_rectified) |
            Q(amount_due__gte=amount_from, state__in=status_due)
        )

    # scenario - both filters set
    if amount_from and amount_to:
        return (
            Q(amount_notified__gte=amount_from, amount_notified__lte=amount_to, state__in=status_notified) |
            Q(amount_rectified__gte=amount_from, amount_rectified__lte=amount_to, state__in=status_rectified) |
            Q(amount_due__gte=amount_from, amount_due__lte=amount_to, state__in=status_due)
        )
